//
//  Salsa Core
//  TEST: Pluggable Protocol Handler for "fhttp"
//  Version 0.00.959
//
//  This XPCOM Module sets up an "fhttp" protocol within the XPCOM network
//  layer. It's a testbed for rewriting data passed through via another,
//  normal network channel.  Eventually, this code will be repurposed to
//  overriding the regular "http" protocol, when SalsaCoreEngine determines
//  the URI of the request points to a page we're going to need to do
//  some magic on.
//
//  Copyright 2004-5, Timothy Fries. All Rights Reserved.
//  This code may be reused freely, without restriction. If you use large
//  portions of this code in another project, attribution is appreciated.
//


// XPCOM constant definitions
const kSCHEME = "fhttp";
const kPROTOCOL_NAME = "Filtered HTTP Protocol";
const kPROTOCOL_CONTRACTID = "@mozilla.org/network/protocol;1?name=" + kSCHEME;
const kPROTOCOL_CID = Components.ID("63e4138b-6d88-4f7d-b46a-b5072460c52b");

const kSIMPLEURI_CONTRACTID = "@mozilla.org/network/simple-uri;1";
const kIOSERVICE_CONTRACTID = "@mozilla.org/network/io-service;1";
const nsISupports = Components.interfaces.nsISupports;
const nsIIOService = Components.interfaces.nsIIOService;
const nsIProtocolHandler = Components.interfaces.nsIProtocolHandler;
const nsIRequest = Components.interfaces.nsIRequest;
const nsIChannel = Components.interfaces.nsIChannel;
const nsIStreamListener = Components.interfaces.nsIStreamListener;
const nsIRequestObserver = Components.interfaces.nsIRequestObserver;
const nsIURI = Components.interfaces.nsIURI;

// Protocol definition
function Protocol()
{
}

Protocol.prototype =
{
   QueryInterface: function(iid)
   {
      if (!iid.equals(nsIProtocolHandler) &&
          !iid.equals(nsISupports))
         throw Components.results.NS_ERROR_NO_INTERFACE;
      return this;
   },

   scheme: kSCHEME,
   defaultPort: -1, 
   protocolFlags: nsIProtocolHandler.URI_NORELATIVE |
                  nsIProtocolHandler.URI_NOAUTH,

   allowPort: function(port, scheme)
   {
      return false;
   },

   newURI: function(spec, charset, baseURI)
   {
      var uri = Components.classes[kSIMPLEURI_CONTRACTID].createInstance(nsIURI);
      uri.spec = spec;
      return uri;
   },

   newChannel: function(input_uri)
   {
      // here goes the code that should be run when the protocol gets used.

      // input_uri is a nsIURI, so get a string from it using .spec
      var mySearchTerm = input_uri.spec;

      // strip away the kSCHEME: part
      mySearchTerm = mySearchTerm.substring(mySearchTerm.indexOf(":")+1, mySearchTerm.length);
      //mySearchTerm = encodeURI(mySearchTerm);

/*
      var finalURL = "";
      var defaultSearchURL = "http://www.google.com/search?q=";

      try {
         var prefService = Components.classes["@mozilla.org/preferences-service;1"]
               .getService(Components.interfaces.nsIPrefService);
         var prefBranch = prefService.getBranch(null);
         defaultSearchURL = prefBranch.getComplexValue("browser.search.defaulturl",
               Components.interfaces.nsIPrefLocalizedString).data;
         var searchDS = Components.classes["@mozilla.org/rdf/datasource;1?name=internetsearch"]
               .getService(Components.interfaces.nsIInternetSearchService);

         var searchEngineURI = prefBranch.getCharPref("browser.search.defaultengine");
         if (searchEngineURI) {
            var searchURL = searchDS.GetInternetSearchURL(searchEngineURI, mySearchTerm, 0, 0, {value:0});
            if (searchURL)
               defaultSearchURL = searchURL;
         }
      }
      catch (e) {
      }

      finalURL = defaultSearchURL + mySearchTerm;
*/

      // create dummy nsIChannel instance
      var ios = Components.classes[kIOSERVICE_CONTRACTID]
            .getService(nsIIOService);

      //return ios.newChannel("javascript:document.location.href='"+finalURL+"'", null, null);
      //return ios.newChannel("about:plugins", null, null);
      //return ios.newChannel("http:"+mySearchTerm, null, null);
      //return ios.newChannel("data:text/plain,url="+mySearchTerm, null, null);
      //if (sdata!="") {
      //   mySearchTerm = encodeURI(sdata);
      //}
      var rChan = ios.newChannel("data:text/plain,url="+mySearchTerm, null, null);

      return new ProtocolChannel(input_uri, rChan);
   }
};

// ProtocolChannel definition
function ProtocolChannel(reqUri, wChan)
{
   this.requestUri = reqUri;
   this.wrappedChannel = wChan;

   wChan.originalURI = reqUri;
}

ProtocolChannel.prototype =
{
   QueryInterface: function(iid)
   {
      if (!iid.equals(nsIChannel) &&
          !iid.equals(nsIRequest) &&
          !iid.equals(nsISupports))
         throw Components.results.NS_ERROR_NO_INTERFACE;
      return this;
   },

   get loadFlags() /* nsIRequest */
   {
      return this.wrappedChannel.loadFlags;
   },
   set loadFlags(value) /* nsIRequest */
   {
      return this.wrappedChannel.loadFlags = value;
   },

   get loadGroup() /* nsIRequest */
   {
      return this.wrappedChannel.loadGroup;
   },
   set loadGroup(value) /* nsIRequest */
   {
      return this.wrappedChannel.loadGroup = value;
   },

   get name() /* nsIRequest */
   {
      //return this.wrappedChannel.name;
      return this.requestUri.spec;
   },

   get status() /* nsIRequest */
   {
      return this.wrappedChannel.status;
   },

   cancel: function(status) /* nsIRequest */
   {
      return this.wrappedChannel.cancel(status);
   },

   isPending: function() /* nsIRequest */
   {
      return this.wrappedChannel.isPending();
   },

   resume: function() /* nsIRequest */
   {
      return this.wrappedChannel.resume();
   },

   suspend: function() /* nsIRequest */
   {
      return this.wrappedChannel.suspend();
   },

   get contentCharset() /* nsIChannel */
   {
      return this.wrappedChannel.contentCharset;
   },
   set contentCharset(value) /* nsIChannel */
   {
      return this.wrappedChannel.contentCharset = value;
   },

   get contentLength() /* nsIChannel */
   {
      return this.wrappedChannel.contentLength;
   },
   set contentLength(value) /* nsIChannel */
   {
      return this.wrappedChannel.contentLength = value;
   },

   get contentType() /* nsIChannel */
   {
      return this.wrappedChannel.contentType;
   },
   set contentType(value) /* nsIChannel */
   {
      return this.wrappedChannel.contentType = value;
   },

   get notificationCallbacks() /* nsIChannel */
   {
      return this.wrappedChannel.notificationCallbacks;
   },
   set notificationCallbacks(value) /* nsIChannel */
   {
      return this.wrappedChannel.notificationCallbacks = value;
   },

   get originalURI() /* nsIChannel */
   {
      //return this.wrappedChannel.originalURI;
      return this.requestUri;
   },
   set originalURI(value) /* nsIChannel */
   {
      //return this.wrappedChannel.originalURI = value;
   },

   get owner() /* nsIChannel */
   {
      return this.wrappedChannel.owner;
   },
   set owner(value) /* nsIChannel */
   {
      return this.wrappedChannel.owner = value;
   },

   get securityInfo() /* nsIChannel */
   {
      return this.wrappedChannel.securityInfo;
   },

   get URI() /* nsIChannel */
   {
      return this.wrappedChannel.URI;
   },

   asyncOpen: function(listener, context) /* nsIChannel */
   {
      //throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
      //return this.wrappedChannel.asyncOpen(listener, context);

      var xlistener = new ProtocolListener(listener);
      return this.wrappedChannel.asyncOpen(xlistener, context);
   },

   open: function() /* nsIChannel */
   {
      //return this.wrappedChannel.open();
      throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
   },

   requestUri: null,
   wrappedChannel: null,
};

// ProtocolListener definition
function ProtocolListener(wListener)
{
   this.wrappedListener = wListener;
}

ProtocolListener.prototype =
{
   QueryInterface: function(iid)
   {
      if (!iid.equals(nsIChannel) &&
          !iid.equals(nsIRequestObserver) &&
          !iid.equals(nsIStreamListener) &&
          !iid.equals(nsISupports))
         throw Components.results.NS_ERROR_NO_INTERFACE;
      return this;
   },

   onStartRequest: function(request, context) /* nsIRequestObserver */
   {
      return this.wrappedListener.onStartRequest(request, context);
   },

   onStopRequest: function(request, context, statusCode) /* nsIRequestObserver */
   {
      return this.wrappedListener.onStopRequest(request, context, statusCode);
   },

   onDataAvailable: function(request, context, inputStream, offset, count) /* nsIStreamListener */
   {
      // XXX: This may be slow -- test me more. Sources say this should be in C++ for speed, but I'm
      //      guessing it won't be necessary as long as we restrict our hijacking of
      //      the protocol to only the requests we really need to work with.
      try {
         var scriptablestream = Components.classes["@mozilla.org/scriptableinputstream;1"]
               .createInstance(Components.interfaces.nsIScriptableInputStream);
         scriptablestream.init(inputStream);
         var data = scriptablestream.read(count);

         // TODO: do something with "data" here, such as pass it through an
         //       HTMLStreamParser-based processing engine.
         //data = data.replace(/g/gi, "G");

         var xstream = Components.classes["@mozilla.org/io/string-input-stream;1"]
               .createInstance(Components.interfaces.nsIStringInputStream);
         xstream.setData(data, data.length);
         return this.wrappedListener.onDataAvailable(request, context, xstream, 0, data.length);
      }
      catch (e) {
         // XXX: Remember to remove this catch -- if something in here fails,
         //      we'll eventually want that exception to escape out to the network library.
         //return this.wrappedListener.onDataAvailable(request, context, inputStream, offset, count);
      }
   },

   wrappedListener: null,
};

// ProtocolFactory definition
var ProtocolFactory = new Object();

ProtocolFactory.createInstance = function(outer, iid)
{
   if (outer != null)
      throw Components.results.NS_ERROR_NO_AGGREGATION;

   if (!iid.equals(nsIProtocolHandler) && !iid.equals(nsISupports))
      throw Components.results.NS_ERROR_NO_INTERFACE;

   return new Protocol();
}

// ModuleObject definition
var ModuleObject = new Object();

ModuleObject.registerSelf = function(compMgr, fileSpec, location, type)
{
   compMgr = compMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
/*
   var httpCID = compMgr.contractIDToCID("@mozilla.org/network/protocol;1?name=http");
   compMgr.unregisterFactory(httpCID,null); // need nsIFactory instead of null!
   compMgr.registerFactoryLocation(kPROTOCOL_CID,
                                   kPROTOCOL_NAME,
                                   "@mozilla.org/network/protocol;1?name=http",
                                   fileSpec,
                                   location,
                                   type);
*/
   compMgr.registerFactoryLocation(kPROTOCOL_CID,
                                   kPROTOCOL_NAME,
                                   kPROTOCOL_CONTRACTID,
                                   fileSpec,
                                   location,
                                   type);
}

ModuleObject.getClassObject = function(compMgr, cid, iid)
{
   if (!cid.equals(kPROTOCOL_CID))
      throw Components.results.NS_ERROR_NO_INTERFACE;

   if (!iid.equals(Components.interfaces.nsIFactory))
      throw Components.results.NS_ERROR_NOT_IMPLEMENTED;

   return ProtocolFactory;
}

ModuleObject.canUnload = function(compMgr)
{
   return true;
}

function NSGetModule() {
   return ModuleObject;
}
