//
//  Salsa Core
//  HTML Stream Parser Object
//  Version 0.00.959
//
//  This object consumes a stream of text and performs basic
//  tokenization of it as an HTML-style document, which is a considerably
//  looser syntax than XML.
//
//  Copyright 2004-5, Timothy Fries. All Rights Reserved.
//  This code may be reused freely, without restriction. If you use large
//  portions of this code in another project, attribution is appreciated.
//

function HTMLStreamParser()
{
   this.resetParser();
}

HTMLStreamParser.prototype =
{
   TEXTSECTION: 1,
   BEFORETAGNAME: 2,
   TAGNAME: 3,
   BEFOREATTRIBUTENAME: 4,
   ATTRIBUTENAME: 5,
   BEFOREEQUALSIGN: 6,
   BEFOREATTRIBUTEVALUE: 7,
   ATTRIBUTEVALUE: 8,
   BEFOREENDTAG: 9,

   IMPLICIT_CLOSE_ELEMENTS: new Array("p","li"),
   NO_CLOSE_ELEMENTS: new Array("img"),

   onevent: null,

   parseStack: null,
   currentState: 0,

   attributeQuoteChar: null,
   tagSelfClose: false,
   isCloseTag: false,
   textSectionBuffer: "",
   currentTagBuffer: "",
   tagNameBuffer: "",
   attributeNameBuffer: "",
   attributeValueBuffer: null,
   attributeBundleBuffer: null,

   resetParser: function()
   {
      this.parseStack = new Array();
      this.currentState = this.TEXTSECTION;
      this.attributeQuoteChar = null;
      this.tagSelfClose = false;
      this.isCloseTag = false;
      this.textSectionBuffer = "";
      this.currentTagBuffer = "";
      this.tagNameBuffer = "";
      this.attributeNameBuffer = "";
      this.attributeValueBuffer = null;
      this.attributeBundleBuffer = null;
   }, 

   parseData: function(inData)
   {
      for (var i=0; i<inData.length; i++)
      {
         var thisChar = inData[i];
         switch (this.currentState)
         {
            case this.TEXTSECTION:
               if ( thisChar == '<' ) {
                  this.currentState = this.BEFORETAGNAME;
                  this.attributeBundleBuffer = new Object();
                  this.tagNameBuffer = "";
                  this.currentTagBuffer = "<";
                  this.isCloseTag = false;
               } else {
                  this.textSectionBuffer += thisChar;
               }
               break;
            case this.BEFORETAGNAME:
               this.currentTagBuffer += thisChar;
               if ( thisChar == '>' ) {
                  this.currentState = this.TEXTSECTION;
                  this.abandonTag();
               } else if ( (thisChar >= 'A' && thisChar <= 'Z') || (thisChar >= 'a' && thisChar <= 'z') ) {
                  this.currentState = this.TAGNAME;
                  this.tagNameBuffer = thisChar;
               } else if ( thisChar == '/' ) {
                  this.isCloseTag = true;
               }
               break;
            case this.TAGNAME:
               this.currentTagBuffer += thisChar;
               if ( thisChar == ' ' ) {
                  this.currentState = this.BEFOREATTRIBUTENAME;
               } else if ( thisChar == '/' ) {
                  this.tagSelfClose = true;
                  this.currentState = this.BEFOREENDTAG;
               } else if ( thisChar == '>' ) {
                  this.currentState = this.TEXTSECTION;
                  this.finalizeTag();
               } else {
                  this.tagNameBuffer += thisChar;
               }
               break;
            case this.BEFOREENDTAG:
               this.currentTagBuffer += thisChar;
               if ( thisChar == '>' ) {
                  this.currentState = this.TEXTSECTION;
                  this.finalizeTag();
               } else if ( thisChar != ' ' && thisChar != '\n' && thisChar != '\r' ) {
                  this.currentState = this.TEXTSECTION;
                  this.abandonTag();
               }
               break;
            case this.BEFOREATTRIBUTENAME:
               this.currentTagBuffer += thisChar;
               if ( thisChar == '>' ) {
                  this.currentState = this.TEXTSECTION;
                  this.finalizeTag();
               } else if ( (thisChar >= 'A' && thisChar <= 'Z') || (thisChar >= 'a' && thisChar <= 'z') ) {
                  this.currentState = this.ATTRIBUTENAME;
                  this.attributeNameBuffer = thisChar;
                  this.attributeValueBuffer = null;
               }
               break;
            case this.ATTRIBUTENAME:
               this.currentTagBuffer += thisChar;
               if ( thisChar == '>' ) {
                  this.currentState = this.TEXTSECTION;
                  this.finalizeValuelessAttribute();
                  this.finalizeTag();
               } else if ( thisChar == ' ' ) {
                  this.currentState = this.BEFOREEQUALSIGN;
               } else if ( thisChar == '=' ) {
                  this.currentState = this.BEFOREATTRIBUTEVALUE;
               } else if ( thisChar == '/' ) {
                  this.currentState = this.BEFOREENDTAG;
                  this.tagSelfClose = true;
                  this.finalizeValuelessAttribute();
               } else {
                  this.attributeNameBuffer += thisChar;
               }
               break;
            case this.BEFOREEQUALSIGN:
               this.currentTagBuffer += thisChar;
               if ( thisChar == '>' ) {
                  this.currentState = this.TEXTSECTION;
                  this.finalizeValuelessAttribute();
                  this.finalizeTag();
               } else if ( thisChar == '=' ) {
                  this.currentState = this.BEFOREATTRIBUTEVALUE;
               } else if ( thisChar == '/' ) {
                  this.currentState = this.BEFOREENDTAG;
                  this.tagSelfClose = true;
                  this.finalizeValuelessAttribute();
               } else if ( (thisChar >= 'A' && thisChar <= 'Z') || (thisChar >= 'a' && thisChar <= 'z') ) {
                  this.currentState = this.ATTRIBUTENAME;
                  this.finalizeValuelessAttribute();
                  this.attributeNameBuffer = thisChar;
                  this.attributeValueBuffer = null;
               }
               break;
            case this.BEFOREATTRIBUTEVALUE:
               this.currentTagBuffer += thisChar;
               if ( thisChar == '\'' || thisChar == '"' ) {
                  this.currentState = this.ATTRIBUTEVALUE;
                  this.attributeValueBuffer = "";
                  this.attributeQuoteChar = thisChar;
               } else if ( thisChar == '>' ) {
                  this.currentState = this.TEXTSECTION;
                  this.finalizeValuelessAttribute();
                  this.finalizeTag();
               } else if ( thisChar == '/' ) {
                  this.currentState = this.BEFOREENDTAG;
                  this.tagSelfClose = true;
                  this.finalizeValuelessAttribute();
               } else if ( thisChar != ' ' && thisChar != '\n' && thisChar != '\r' ) {
                  this.currentState = this.ATTRIBUTEVALUE;
                  this.attributeValueBuffer = thisChar;
                  this.attributeQuoteChar = null;
               }
               break;
            case this.ATTRIBUTEVALUE:
               this.currentTagBuffer += thisChar;
               if ( (this.attributeQuoteChar == null && (thisChar == ' ' || thisChar == '\n' || thisChar == '\r')) ||
                    (this.attributeQuoteChar != null && thisChar == this.attributeQuoteChar) ) {
                  this.currentState = this.BEFOREATTRIBUTENAME;
                  this.finalizeValuedAttribute();
               } else if ( this.attributeQuoteChar == null && thisChar == '/' ) {
                  this.currentState = this.BEFOREENDTAG;
                  this.finalizeValuedAttribute();
                  this.tagSelfClose = true;
               } else if ( this.attributeQuoteChar == null && thisChar == '>' ) {
                  this.currentState = this.TEXTSECTION;
                  this.finalizeValuedAttribute();
                  this.finalizeTag();
               } else {
                  this.attributeValueBuffer += thisChar;
               }
               break;
         }
      }
   },

   parseEndOfData: function()
   {
      switch ( this.currentState )
      {
         case this.TEXTSECTION:
            this.finalizeTextSection();
            break;
         case this.BEFORETAGNAME:
         case this.TAGNAME:
         case this.BEFOREATTRIBUTENAME:
         case this.ATTRIBUTENAME:
         case this.BEFOREEQUALSIGN:
         case this.BEFOREATTRIBUTEVALUE:
         case this.ATTRIBUTEVALUE:
         case this.BEFOREENDTAG:
            this.abandonTag();
            this.finalizeTextSection();
            break;
      }
      this.raiseEvent("end", null, null);
   },

   abandonTag: function()
   {
      this.textSectionBuffer += this.currentTagBuffer;
      this.currentTagBuffer = "";
      this.attributeBundleBuffer = null;
   },

   finalizeTextSection: function()
   {
      if ( this.textSectionBuffer.length > 0 ) {
         this.raiseEvent("text", this.textSectionBuffer, null);
      }
      this.textSectionBuffer = "";
   },

   finalizeTag: function()
   {
      this.finalizeTextSection();

      this.attributeBundleBuffer["_tagname"] = this.tagNameBuffer;
      var lcTagName = this.tagNameBuffer.toLowerCase();
      for ( var i=0; i<this.NO_CLOSE_ELEMENTS.length; i++ ) {
         if ( this.NO_CLOSE_ELEMENTS[i] == lcTagName ) {
            this.tagSelfClose = true;
         }
      }

      if ( this.isCloseTag ) {
         this.raiseEvent("close", this.tagNameBuffer, null);
      } else {
         if ( this.tagSelfClose ) {
            this.raiseEvent("selfclose", this.tagNameBuffer, this.attributeBundleBuffer);
         } else {
            this.raiseEvent("open", this.tagNameBuffer, this.attributeBundleBuffer);
         }
      }

      this.currentTagBuffer = "";
      this.tagSelfClose = false;
      this.isCloseTag = false;
   },

   finalizeValuelessAttribute: function()
   {
      this.attributeQuoteChar = null;
      this.attributeBundleBuffer[ this.attributeNameBuffer ] = true;
      this.dumpAttributeData();
   },

   finalizeValuedAttribute: function()
   {
      this.attributeQuoteChar = null;
      this.attributeBundleBuffer[ this.attributeNameBuffer ] = this.attributeValueBuffer;
      this.dumpAttributeData();
   },

   dumpAttributeData: function()
   {
      this.attributeNameBuffer = "";
      this.attributeValueBuffer = null;
   },

   raiseEvent: function(name, data, attrib)
   {
      if ( typeof(this.onparseevent) == "function" ) {
         this.onparseevent(name, data, attrib);
      }
   }
};
