//
//  Salsa Core
//  HTML Stream Tree Stack Manager Object
//  Version 0.00.959
//
//  This object consumes tokens as outputted by HTMLStreamParser and
//  manages the details of maintaining the stack of which tags are
//  contained by which other tags; exposing events that can be used
//  to monitor the document structure as it streams through.
//
//  Copyright 2004-5, Timothy Fries. All Rights Reserved.
//  This code may be reused freely, without restriction. If you use large
//  portions of this code in another project, attribution is appreciated.
//

function HTMLStreamTreeStack()
{
   this.parseStack = new Array();
}

HTMLStreamTreeStack.prototype =
{
   IMPLICIT_CLOSE_ELEMENTS: new Array("p", "li", "td"),

   parseStack: null,

   ontreeevent: null,

   parserDelegate: function()
   {
      var xthis = this;
      return function(name, data, attrib) {
         xthis.parserEvent(name, data, attrib);
      };
   },

   parserEvent: function(name, data, attrib)
   {
      switch(name)
      {
         case "text":
            this.textBlock(data);
            break;
         case "open":
            this.tagOpen(data, attrib);
            break;
         case "close":
            this.tagClose(data, true);
            break;
         case "selfclose":
            this.tagOpenAndClose(data, attrib);
            break;
         case "end":
            this.finalizeTree();
            break;
      }
   },

   textBlock: function(data)
   {
      this.exportEvent("text", data);
   },

   tagOpen: function(data, attrib)
   {
      //TODO: support implicitly closing elements
      this.parseStack.push( attrib );
      this.exportEvent("open", null);
   },

   tagClose: function(data, explicit)
   {
      var closeIndex = -1;
      var len = this.parseStack.length;
      for (var m=0; m<len; m++) {
         if (this.parseStack[len-m-1]["_tagname"] == data) {
            closeIndex = len - m - 1;
            m = len;
         }
      }
      if ( closeIndex >= 0 ) {
         while ( this.parseStack.length > closeIndex+1 ) {
            this.tagClose(this.parseStack[this.parseStack.length-1]["_tagname"], false);
         }
         this.exportEvent("close", { isExplicit: explicit });
         this.parseStack.pop();
      } else {
         this.tagMismatchClose(data);
      }
   },

   tagMismatchClose: function(data)
   {
      this.exportEvent("badclose", null);
   },

   tagOpenAndClose: function(data, attrib)
   {
      this.parseStack.push( attrib );
      this.exportEvent("selfclose", null);
      this.parseStack.pop();
   },

   finalizeTree: function()
   {
      while ( this.parseStack.length > 0 ) {
         this.tagClose(this.parseStack[this.parseStack.length-1]["_tagname"], false);
      }
      this.exportEvent("end", null);
   },

   exportEvent: function(ename, data)
   {
      if ( typeof(this.ontreeevent) == "function" ) {
         this.ontreeevent(this, ename, data);
      }
   },

   _stackAsString: function()
   {
      var res = new Array();
      for (var i=0; i<this.parseStack.length; i++) {
         res.push(this.parseStack[i]["_tagname"]);
      }
      return "/"+res.join("/");
   },
};
